from time import time
from typing import Union, Dict

from py_jwt_verifier.jwk import JWK
from py_jwt_verifier.jwt import JWT
from py_jwt_verifier.jws import JWS
from py_jwt_verifier.exceptions import PyJwtException


__all__ = [
    "PyJwtVerifier",
]


class PyJwtVerifier:

    def __init__(
        self,
        jwt: str,
        alg: str = None,
        cid: str = None,
        aud: str = None,
        iss: str = None,
        auto_verify=True,
        check_expiry=True,
        cache_enabled=True,
        cache_lifetime=1,
        cache_store="sqlite",
        oidc_metadata_url=None,
        custom_claims: dict = None,
    ) -> None:
        self.jwt = JWT.from_encoded(jwt)
        if check_expiry:
            self.is_expired()

        self.alg = alg
        self.cid = cid
        self.aud = aud
        self.iss = iss
        if not custom_claims:
            custom_claims = {}
        self.custom_claims = custom_claims

        self.jwk = JWK(
            cache_enabled=cache_enabled,
            cache_lifetime=cache_lifetime,
            cache_store=cache_store,
            oidc_metadata_url=oidc_metadata_url,
        )

        self.jws = JWS(
            jwt=self.jwt,
            jwk=self.jwk,
        )

        if auto_verify is True:
            self.verify()

    def is_expired(self) -> None:
        exp = self.jwt.payload.get("exp")
        if exp:
            time_now = int(time())

            if time_now >= exp:
                raise PyJwtException("exp")

    def verify(self, get_payload=False) -> Union[dict, None]:
        # Validate claims.
        if self.cid:
            if self.cid != self.jwt.payload.get("cid"):
                raise PyJwtException("cid")
        if self.aud:
            if self.aud != self.jwt.payload.get("aud"):
                raise PyJwtException("aud")
        if self.iss:
            if self.iss != self.jwt.payload.get("iss"):
                raise PyJwtException("iss")
        if self.custom_claims:
            if not isinstance(self.custom_claims, dict):
                raise PyJwtException("custom_claims_type")
            # k - claim name, v - claim value
            for k, v in self.custom_claims.items():
                if v != self.jwt.payload.get(k):
                    raise PyJwtException("custom_claim")

        # Validate the signature.
        self.jws.verify_signature()

        # Optional output.
        if get_payload is True:
            return self.return_payload()

    def return_payload(self) -> Dict[str, dict]:
        return {"header": self.jwt.header, "payload": self.jwt.payload}



from base64 import b64decode, urlsafe_b64decode
from json import loads
from struct import unpack
from typing import Tuple, List, Dict


__all__ = ["b64_decode"]


def b64_decode(jwt: List[str]) -> List[Dict[str, dict]]:
    """
    :param jwt: JWT string split into a list of exactly 3 strings.
    :returns: List of 2 dicts: [{header}, {payload}]
    """
    # Removing signature [2]
    decoded_jwt = []
    for part in jwt[0:2]:
        part = part.replace("_", "/")
        part = part.replace("-", "+")
        # Add padding
        part += "=" * (4 - len(part) % 4)
        decoded_jwt.append(loads(b64decode(part)))
    return decoded_jwt


def _int_arr2_long(arr: Tuple[int]) -> int:
    return int("".join(["%02x" % byte for byte in arr]), 16)


def base64_to_long(data) -> int:
    if isinstance(data, str):
        data = data.encode("ascii")
    _d = urlsafe_b64decode(bytes(data) + b"==")
    return _int_arr2_long(unpack("%sB" % len(_d), _d))



from typing import Union, List, Dict, Any

from py_jwt_verifier.exceptions import PyJwtException
from py_jwt_verifier import utils


__all__ = ["JWT"]

PY_JWT_SEPARATOR = "."
PY_JWT_EXPECTED_LEN = 3
PY_JWT_HEADER_IDX = 0
PY_JWT_PAYLOAD_IDX = 1
PY_JWT_SIGNATURE_IDX = 2


class JWT:

    __slots__ = (
        "encoded",
        "decoded",
        "header",
        "payload",
        "message",
        "signature",
    )

    @classmethod
    def from_encoded(cls, encoded: str) -> "JWT":
        encoded_split = encoded.split(PY_JWT_SEPARATOR)
        # Ensure the format is valid by checking that splitting the
        # encoded str by "." will result in a list with len == {PY_JWT_EXPECTED_LEN}.
        # Will raise PyJwtException("len") when lentgth isn't the exepcted one.
        if len(encoded_split) != PY_JWT_EXPECTED_LEN:
            raise PyJwtException("len")

        # Set the header, payload, message and signature.
        decoded = utils.b64_decode(encoded_split)
        header, payload = (
            decoded[PY_JWT_HEADER_IDX],
            decoded[PY_JWT_PAYLOAD_IDX],
        )
        message = (
            encoded_split[PY_JWT_HEADER_IDX].encode("ascii")
            + PY_JWT_SEPARATOR.encode("ascii")
            + encoded_split[PY_JWT_PAYLOAD_IDX].encode("ascii")
        )
        signature = encoded_split[PY_JWT_SIGNATURE_IDX]

        return JWT(
            encoded=encoded,
            decoded=decoded,
            header=header,
            payload=payload,
            message=message,
            signature=signature,
        )

    def __init__(
        self,
        encoded: str,
        decoded: List[Dict[str, Any]],
        header: Dict[str, Any],
        payload: Dict[str, Any],
        message: bytes,
        signature: str,
    ) -> None:
        self.encoded = encoded
        self.decoded = decoded
        self.header = header
        self.payload = payload
        self.message = message
        self.signature = signature



__all__ = [
    "JWSVerifier",
]


class JWSVerifier:
    """
    Provide skeleton for JWS Verifier implementations.
    """

    def verify_signature(self):
        raise NotImplementedError("All sub-classes must implement this method.")



from py_jwt_verifier.exceptions import PyJwtException
from py_jwt_verifier.jws.base import JWSVerifier
from py_jwt_verifier.jws.rs256 import RSASSA_PKCS1_V1_5_VERIFY
from py_jwt_verifier.jwk import JWK
from py_jwt_verifier.jwt import JWT


__all__ = ["JWS"]


PY_JWT_ALG_RS256 = "rs256"


def _create_signature_verifier(jwt: JWT, jwk: JWK) -> JWSVerifier:
    alg = str(jwt.header.get("alg")).lower()
    if alg == PY_JWT_ALG_RS256:
        return RSASSA_PKCS1_V1_5_VERIFY(jwt, jwk)
    raise PyJwtException("alg")


class JWS:

    def __init__(self, jwt: JWT, jwk: JWK) -> None:
        self._verifier = _create_signature_verifier(jwt, jwk)

    def verify_signature(self):
        self._verifier.verify_signature()



from datetime import timedelta
from json.decoder import JSONDecodeError
from typing import Union, Tuple, Literal

from requests import Session
from requests_cache import (
    CachedSession,
    BaseCache,
    FileCache,
    SQLiteCache,
    RedisCache,
    MongoCache,
    GridFSCache,
    DynamoDbCache,
)
from requests.exceptions import SSLError

from py_jwt_verifier.exceptions import PyJwtException

PY_JWT_CACHE_NAME = "pyw-jwt-verifier"
PY_JWT_DEFAULT_CACHE_STORE = "sqlite"

__all__ = ["JWK"]


class JWK:
    def __init__(
        self,
        cache_enabled: bool = True,
        cache_store: Union[
            Literal[
                "memory",
                "filesystem",
                "sqlite",
                "redis",
                "mongodb",
                "gridfs",
                "dynamodb",
            ],
            BaseCache,
            FileCache,
            SQLiteCache,
            RedisCache,
            MongoCache,
            GridFSCache,
            DynamoDbCache,
        ] = PY_JWT_DEFAULT_CACHE_STORE,
        cache_lifetime: int = 1,
        oidc_metadata_url: Union[str, None] = None,
    ) -> None:
        if cache_lifetime > 30 or cache_lifetime < 1:
            raise PyJwtException("cache-lifetime")
        self.oidc_metadata_url = oidc_metadata_url

        if cache_enabled:
            try:
                self._session = CachedSession(
                    cache_name=PY_JWT_CACHE_NAME,
                    backend=cache_store,
                    expire_after=cache_lifetime,
                )
            except ValueError:
                raise PyJwtException("cache-store")
        else:
            self._session = Session()

    def get_e_n(self, issuer: str, kid: Union[str, None] = None) -> Tuple[str, str]:
        if kid:
            return self._get_e_n_with_kid(issuer, kid)
        return self._get_e_n(issuer)

    def _get_e_n_with_kid(self, issuer: str, kid: str) -> Tuple[str, str]:
        keys_endpoint = self._compute_keys_endpoint(issuer)
        keys = self._get_json_response(keys_endpoint).get("keys", [])
        for key in keys:
            if key.get("kid") == kid:
                return key.get("e"), key.get("n")
        raise PyJwtException("kid")

    def _get_e_n(self, _: str) -> Tuple[str, str]:
        raise NotImplementedError("Not yet suupported.")

    def _compute_keys_endpoint(self, issuer) -> str:
        metadata_url = self.oidc_metadata_url

        if metadata_url is None:
            metadata_url = f"{issuer}/.well-known/openid-configuration"

        # Some issuers end with a trailing "/".
        # Auth0 is one of the idps that set the issuer as such.
        if "//." in metadata_url:
            metadata_url = f"{issuer}.well-known/openid-configuration"

        metadata = self._get_json_response(metadata_url)
        return metadata.get("jwks_uri")

    def _get_json_response(self, url) -> dict:
        try:
            return self._session.get(url).json()
        except SSLError:
            raise PyJwtException("ssl")
        except JSONDecodeError:
            raise PyJwtException("json")



import hashlib

from py_jwt_verifier.exceptions import PyJwtException
from py_jwt_verifier.jwk import JWK
from py_jwt_verifier.jwt import JWT
from py_jwt_verifier.jws.base import JWSVerifier
from py_jwt_verifier.utils import base64_to_long


__all__ = ["RSASSA_PKCS1_V1_5_VERIFY"]


class RSASSA_PKCS1_V1_5_VERIFY(JWSVerifier):
    """
    RFC - JSON Web Signature (JWS) -  RSASSA-PKCS1-V1_5-VERIFY ((n, e), M, S)
    https://tools.ietf.org/html/rfc3447.html#section-8.2.2

    HUMAN-READABLE ARTICLE THAT EXPLAINS THIS PROCESS -
    https://medium.com/@bn121rajesh/rsa-sign-and-verify-using-openssl-behind-the-scene-bf3cac0aade2
    """

    def __init__(self, jwt: JWT, jwk: JWK) -> None:
        self.jwt = jwt
        self.jwk = jwk

    def verify_signature(self) -> None:
        iss, kid = self.jwt.payload.get("iss"), self.jwt.header.get("kid")
        exponent, modulus = self.jwk.get_e_n(iss, kid)
        # Length Checking. len(signature) must equal len(modulus)
        if len(self.jwt.signature) != len(modulus):
            raise PyJwtException("sig")

        # RSAVP1 verification
        s = base64_to_long(self.jwt.signature)
        e = base64_to_long(exponent)
        n = base64_to_long(modulus)
        self._RSASVP1(s, n)

        # Obtain computed hash of header + . + payload
        computed_hash = self._message_hash(self.jwt.message)

        # Obtain signature hash
        public_hash = self._signature_hash(s, e, n)

        if computed_hash != public_hash:
            raise PyJwtException("sig")

    def _RSASVP1(self, s, n) -> None:
        """
        https://tools.ietf.org/html/rfc3447.html#section-5.2.2
        s should be greater than 0 and less than (n-1)
        """
        if s < 0 and s > (n - 1):
            raise PyJwtException("sig")

    def _message_hash(self, message) -> str:
        digest = hashlib.sha256()
        digest.update(message)
        digest = digest.hexdigest()
        return digest[-40:]

    def _signature_hash(self, s: int, e: int, n: int) -> str:
        s_hash = hex(pow(s, e, n))
        return s_hash[-40:]



import platform
import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

urllib3_version = "<2.0"
major, minor, patchlevel = platform.python_version_tuple()
if int(major) == 3 and int(minor) > 7:
    urllib3_version = ">=2.0"

setuptools.setup(
    name="py_jwt_verifier",
    version="0.8.0",
    author="Adrian Lazar",
    author_email="adrian.lazar95@outlook.com",
    description="Python JWT Verfier - Verifies the signature of a digitally signed JWT.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/adrianlzr/py-jwt-verifier",
    keywords="python jwt verifier, jwt verifier, jwt signature verifier, py-jwt-verifier, python-jwt-verifier, python jwt signature verifier",
    packages=setuptools.find_packages(),
    install_requires=[
        "requests >=2.31.0, <2.40.0",
        "requests_cache >=1.2.0, <1.3.0",
        "urllib3%s" % urllib3_version,
    ],
    python_requires=">=3.8.0",
    project_urls={
        "Documentation & Source": "https://github.com/adrianlzr/py-jwt-verifier",
        "Issue Tracker": "https://github.com/adrianlzr/py-jwt-verifier/issues",
    },
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Security :: Cryptography ",
    ],
)



__all__ = [
    "PY_JWT_ERRORS",
    "PY_JWT_DEFAULT_ERROR",
    "PyJwtException",
]


PY_JWT_ERRORS = {
    "alg": "Unsupported algorithm. Supported algorithms: RS256.",
    "len": "JWT length is invalid.",
    "exp": "JWT is expired.",
    "cid": "Invalid Client ID present in the payload.",
    "aud": "Invalid Audience present in the payload.",
    "iss": "Invalid Issuer present in the payload.",
    "custom_claim": "The specified Custom Claim/s does not seem to be valid/exist.",
    "custom_claims_type": "The **custom_claims object must be of type dict!",
    "sig": "JWT Signature is not valid.",
    "kid": """Unable to find a matching kid. Please invalidate the cache as the keys might've been rotated. If the issue persists 
    the JWT is either invalid or in case of Okta: Access Tokens can only be validated locally if they were issued by a Custom Authorization Server.""",
    "ssl": "The SSL Certificate seems to be issued for a different domain or the certificate chain is not present.",
    "cache-store": "Invalid cache store. Please choose between: memory, sqlite, mongo and redis.",
    "cache-lifetime": "Invalid cache lifetime! The lifetime must be higher than 1 and less than 30 days.",
    "json": "Invalid JSON data. Either the /keys endpoint is not reachable or the response data is invalid.",
}

PY_JWT_DEFAULT_ERROR = "General, non-determined error."


class PyJwtException(Exception):
    def __init__(self, arg: str = None) -> None:
        super().__init__(PY_JWT_ERRORS.get(arg, PY_JWT_DEFAULT_ERROR))
